/*
# Copyright (c) 2003-2004 by the University of Southern California.
# All rights reserved.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation in source and binary forms for non-commercial purposes and
# without fee is hereby granted, provided that the above copyright notice
# appear in all copies and that both the copyright notice and this permission
# notice appear in supporting documentation, and that any documentation,
# advertising materials, and other materials related to such distribution and
# use acknowledge that the software was developed by the University of
# Southern California, Information Sciences Institute.  The name of the
# University may not be used to endorse or promote products derived from this
# software without specific prior written permission.
#
# THE UNIVERSITY OF SOUTHERN CALIFORNIA MAKES NO REPRESENTATIONS ABOUT THE
# SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE.  THIS SOFTWARE IS PROVIDED "AS
# IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT
# LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE.
#
# Other copyrights might apply to parts of this software and are so noted when
# applicable.
# ----------------------------------------------------------------------------
#
# Primary author: Purushotham Kamath
#                 Sept  2004
#This material is based upon work supported by the Defense Advanced Research 
#Projects Agency under contract no. N66001-02-1-8939 issued by the Space and 
#Naval Warfare Systems Center (SPAWAR).  Any opinions, findings, and conclusions 
#or recommendations expressed in this material are those of the author(s) and 
#do not necessarily reflect the views of the Defense Advanced Research Projects
#Agency, SPAWAR, or the U.S. Government.
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>


#include <packet.h>
#include <eventhandler.h>
#include <queue.h>
#include <scheduler.h>
#include <state.h>
#include <rng.h>
#include <trgen.h>
#include <trdump.h>
#include <txsched.h>
#include <intransit.h>
#include <codeset.h>
#include <debug.h>
#include <receiver.h>
#include <stateest.h>
#include <state.h>
#include <statshandler.h>

extern Scheduler *sched;
extern TrGen * trgen;
extern InTransit *it;
extern StateEstimator *stateest;
extern Codeset *codeset;
extern int steadystate;

StatsHandler::StatsHandler(double n, double a, double i, double d)
{

	A = a;
	N = n;
	interval = i;
	distance = d;
	totaltc=0;
	totalsc=0;
	totalonline=0;
	totalscreadings=0;
	totaltcreadings=0;
	totalonlinereadings=0;
	oldstate = NULL;
	totaltsc=0.0;
	totaltscreadings =0;

	debug = debugon("statshandler");
}

StatsHandler::~StatsHandler()
{

}

void StatsHandler::handleevent(Packet *p)
{


	dprintf("statshandler.cc: StatsHandler::calcpcc(): Calculations complete\n"); 		
	
	calcnumonline();
	calcspatialcoherence();
	calctemporalcoherence();
	
	if(trgen->done != 1)
		sched->scheduleevent(this, NULL, interval);	
}


void StatsHandler::process(Packet *p)
{


}


void StatsHandler::dumpstats()
{
	double avgsc = 0;
	double avgtc = 0;
	double avgonline = 0;
	avgsc = totalsc/(double)totalscreadings;
	avgtc = totaltc/(double)totaltcreadings;
	avgonline = totalonline/(double)totalonlinereadings;

	double normpktthput2 = (avgonline/N)*( 1 - avgper);
	double avgtsc = (double) totaltsc/(double)totaltscreadings;


	printf("%d %d %.4f %.4f %.4f %.4f %.4f %.4f %d %.6f %.6f %.6f\n", totalpkts, noerrorpkts, normoffload, normpktthput, normpktthput2, normbytethput, avgonline, avgper, droppedpkts, avgtc, avgsc, avgtsc);
	//printf("%.12f\n", sched->currenttime);
}



double StatsHandler::calcl2(int *s1, int *s2)
{

//Calculates the difference between two state variables.
//Uses an L2 metric (an RMS metric) as the measure of difference between the states.
	
	double mtotaldiff = 0.0;
	double totaldiff = 0.0;

	for (int count1 =0; count1 < A; count1++)
	{
		for (int count =(int)N - 1; count >=0 ; count--)
		{
			int diff = s1[(int)(count1*N + count)] - s2[(int)(count1*N + count)];
			totaldiff+=pow (diff, 2);
			mtotaldiff = totaldiff/(A*N);
		}
	}
	
	double l2diff = pow(mtotaldiff, 0.5);	
	return l2diff;


}


double StatsHandler::calcpcc(int *s1, int *s2)
{


//Pearson's correlation coefficient distance metric
	double xi =0;
	double yi =0;
	for (int count1 =0; count1 < A; count1++)
	{
		for (int count =(int)N - 1; count >=0 ; count--)
		{
			xi+=s1[(int ) (count1*(N) +count)];
			yi+=s2[(int)(count1*(N) + count)];
			//printf("serror: %.6f\n", se);
		}
	}
	double xmean = xi/((A)*(N));
	double ymean = yi/((A)*(N));

	dprintf("statshandler.cc: StatsHandler::calcpcc(): xmean: %.6f\n", xmean); 		
	dprintf("statshandler.cc: StatsHandler::calcpcc(): ymean: %.6f\n", ymean); 		
	
	double totalxydev =0;	
	double totalxdev2 =0;	
	double totalydev2 =0;	
	for (int count1 =0; count1 < A; count1++)
	{
		for (int count =(int)N - 1; count >=0 ; count--)
		{
			double xdev = s1[(int)(count1*(N) +count)] - xmean;
			double ydev = s2[(int)(count1*(N) +count)] - ymean;
			totalxydev+= xdev*ydev; 
			totalxdev2+=pow(xdev, 2);
			totalydev2+=pow(ydev, 2);
		}
	}
	dprintf("statshandler.cc: StatsHandler::calcpcc(): totalxydev: %.6f\n", totalxydev); 		
	dprintf("statshandler.cc: StatsHandler::calcpcc(): totalxdev2: %.6f\n", totalxdev2); 		
	dprintf("statshandler.cc: StatsHandler::calcpcc(): totalydev2: %.6f\n", totalydev2); 		
	double pcc = 0;
	if ((totalxdev2 !=0 ) && (totalydev2 !=0))
		pcc = totalxydev/(pow(totalxdev2, 0.5) * pow(totalydev2, 0.5));
	else 
	 	pcc =  0.0;	

	dprintf("statshandler.cc: StatsHandler::calcpcc(): pcc: %.6f\n", pcc); 		
	return pcc;

}


void StatsHandler::calcspatialcoherence()
{

	dtprintf("statshandler.cc: StatsHandler::calcspatial(): Calculating spatial coherence.\n");
	double p1=0;
	double p2=distance;

	dtprintf("statshandler.cc: Statshandler::calcspatialcoherence: calculating state at : %.6f\n", p1);
	int * s1 = stateest->calcstate(p1);
	dtprintf("statshandler.cc: Statshandler::calcspatialcoherence: calculating state at : %.6f\n", p2);
	int * s2 = stateest->calcstate(p2);
	int chipoffset = stateest->calcchipoffset(p1, p2);			
	int * s2t = codeset->LeftShiftCodeword(s2, chipoffset);	
		
		
	double sc = calcpcc(s1, s2t);
	totalsc+=sc; 
	totalscreadings++;

}


void StatsHandler::calcnumonline()
{

//Collect stats about the number of codewords online.
//Use the point (0,0,0) as the reference point
	int numonline = 0;
	Packet *p = it->q->head;
	while(p != NULL )
	{
	
		double packethead = 0.0;
		double packettail = 0.0;
		p->calcboundary(&packethead, &packettail);
		
//Packet does not overlap
		if((packettail > 0 ) && (packethead < 0))
		{
			//printf("%d %.6f %.6f ", p->id, packethead, packettail);
			numonline++;
		}

		p=p->next;
	}
	dprintf("\nstatshandler.cc: Statshandler::handleevenT(): Number on line: %d\n", numonline);
	int t = it->numonline();
	dprintf("statshandler.cc: Statshandler::handleevenT(): Number in transit: %d\n",t);
	totalonline+=numonline;	
	totalonlinereadings++;
}



void StatsHandler::calctemporalcoherence()
{
//Dump state at point (0,0,0) 

	dtprintf("statshandler.cc: StatsHandler::calctemporalcoherence(): Calculating temporal coherence: \n");
	int * s1 = stateest->calcstate(0.0);
	
	if(oldstate == NULL)
		oldstate = s1;

	double tc = calcpcc(s1, oldstate);
	totaltc+=tc;
	totaltcreadings++;

	if((oldstate != NULL) && (oldstate != s1))
		delete oldstate;
	oldstate = s1;
}	

int StatsHandler::dumpstate(int *s1)
{

	dprintf("statshandler.cc:Statshandler::dumpstate(): ");
	for (int count1 =0; count1 < A; count1++)
	{
		for (int count =(int)N - 1; count >=0 ; count--)
		{
			dprintf("%d ", s1[(int) (count1*N + count)]);
		}
	}
	dprintf("\n");
	return 0;
}

void StatsHandler::calcstatediff(Packet *rp)
{

//Calculate the error in state estimation due to temporal and spatial coherence
//State(Est pt, Est time) - State(Merge pt, Merge time)

//Note : Here state is measured at the receive point. This is OK because there is no difference
//between state at the merge point at merge time and state at receive point at receive time.
//This is because there are no arrivals after the coupler.

	double x, t;
	rp->calcboundary(&x, &t);

	dprintf("stathandler.cc: StatsHandler::calcstatediff(): Real state at %.6f\n", x);
	
	int * state = new int [(rp->A) *(rp->N)];
	if(state == NULL)
	{
		printf("Error: statshandler.cc: Memory allocation failed\n");
		exit(-1);
	}
	for (int count = 0; count < (rp->A)*(rp->N); count++)
		state[count] =  0;

	//Only the intransit object contains the actual data
	//So use it to determine state. Don't use the calcstate functions of stateest because they will include rp in the state
	Packet *datap = it->q->head;
	while(datap != NULL )
	{
		if(datap->id == rp->id)
		{
			datap=datap->next;
			continue;
		}
		double head, tail;
		datap->calcboundary(&head, &tail);
		if ((head < x) && (tail > x))
		{	
			int * pstate = datap->calcstate(x, 1);
			for (int count = 0; count < rp->A*rp->N; count++)
			{
				state[count]+= pstate[count];

			}	
			delete pstate;
		}
		datap=datap->next;
	}	

	


//The state at the estimation point will not be bit synchronized with the state at 
//the receiving point
//Hence the phase difference needs to be determined.
	
/*	dtprintf("statshandler.cc: Statshandler::calcpcc(): Finding offset\n");
	double ex = rp->ep;
	//double ex = rp->depx;
	double rx = s->head;
	double chiplength = (1.0/codeset->chiprate) * propSpeed;	
	double offset = ex - rx;

	dtprintf("statshandler.cc: Statshandler::calcpcc(): Offset (chips): %.6f\n", offset);
	int chipoffset =0;
	if (offset >= 0)
		chipoffset = int((offset/chiplength) + 0.5);
	else
		chipoffset = int((offset/chiplength) - 0.5);

	printf("statshandler.cc: Statshandler::calcpcc(): Estimation distance offset (chips): %d\n", chipoffset);
	printf("statshandler.cc: Statshandler::calcpcc(): Tx scheduling offset (chips): %d\n", rp->offset);
	chipoffset = chipoffset % (int) rp->N;	
	if(chipoffset < 0)
		chipoffset = -chipoffset;
*/
	

//	int chipoffset = stateest->calcchipoffset(rp->ep, head);			
//	int * lstate = codeset->LeftShiftCodeword(rp->state, chipoffset);	

	int chipoffset =0;
	chipoffset+=rp->offset;		
	dtprintf("statshandler.cc: Statshandler::calcpcc(): Estimation distance offset (chips): %d\n", chipoffset);
	dtprintf("statshandler.cc: Statshandler::calcpcc(): Estimation pt: %.6f\n", rp->ep);
	//dtprintf("statshandler.cc: Statshandler::calcpcc(): Real pt pt: %.6f\n", head);
	dtprintf("statshandler.cc: Statshandler::calcpcc(): Est pt pt: %.6f\n", rp->ep);
	dtprintf("\n");

	
	int * tstate = codeset->RightShiftCodeword(rp->state, chipoffset);	

	dtprintf("statshandler.cc: StatsHandler::calcpcc(): Real state: \n"); 		
	for (int count1 =0; count1 < rp->A; count1++)
	{
		for (int count =rp->N -1; count >=0; count--)
		{
			dprintf("%d ", state[count1*(rp->N) + count]);
		}
		dprintf("\n");
	}

	dtprintf("statshandler.cc: StatsHandler::calcpcc(): Estimated state: \n"); 		
	for (int count1 =0; count1 < rp->A; count1++)
	{
		for (int count =rp->N -1; count >=0; count--)
		{
			//dprintf("%d ", rp->state[count1*(rp->N) + count]);
			dprintf("%d ", tstate[count1*(rp->N) + count]);
		}
		dprintf("\n");
	}
	dprintf("\n"); 		
/*	dtprintf("statshandler.cc: StatsHandler::calcpcc(): Estimated state shifted: \n"); 		
	for (int count1 =0; count1 < rp->A; count1++)
	{
		for (int count =rp->N -1; count >=0; count--)
		{
			dprintf("%d ", lstate[count1*(rp->N) + count]);
		}
		dprintf("\n");
	}
	dprintf("\n"); 		
*/


//Euclidean distance metric
/*	
	double tsc = stats->calcl2(tstate, rp->state);
*/

	double tsc = calcpcc(tstate, state);
	dprintf("statshandler.cc: StatsHandler::calcpcc(): Pcc: %.6f\n", tsc); 		
	totaltsc+=tsc;
	dtprintf("statshandler.cc: StatsHandler::calcpcc(): totalerror: %.6f\n", totaltsc); 	
	totaltscreadings++;


	delete state;
	//delete tstate;
}
